Attribute VB_Name = "VBwinLAMP"
'    Date: october 2003

'This Visual Basic form module is
'provided as an attachment to "CODiCE iNVERSO", an e-zine
'text about digital cryptography, published on Ondaquadra
'#10 (www.ondaquadra.org) by Zer0 (Massimo De Lirio,
'Italy). The code is "as is", and has not been proved to
'be correct, bug-free or even not potentially dangerous.
'You can distribute it or modify without any limitation
'(I'll be glad to you if you don't remove this intro,
'though...), nor warranty.
'It is not intended to be a definitive program, so please
'don't report me any bug or error you find in the code: I
'will not correct it, this is just part of a sample program
'intended to demonstrate the basics of a multi-precision
'arithmetic library programming in VB.
'If you have any question about, I recommend you to give
'a look at CODiCE iNVERSO (a quite long text, it starts on
'Ondaquadra #03, it's in Italian, though!), you can also
'find in Ondaquadra #10 my e-mail address and my PGP
'public keys.

'KEY 1: RSA 2046/ IDEA 128
'    Key ID: 0xA711DDC9
'    Key fingerprint: 1693 8400 7E5E F1C8  F34E 9970 3892 BD70
'    Created on: 02/06/01 (never expires)

'KEY 2: DH 4095/ DSA 1024/ IDEA 128
'    Key ID: 0x01965550
'    Key fingerprint: 989B 2124 8577 E5B9 3BFE  9B2D B22E 46BB 0196 5550
'    Created on: 19/08/01 (never expires)


'******************************************
'*   Visual Basic L.A.M.P. for Windows    *
'* (Libreria Aritmetica Multi-Precisione) *
'******************************************

Option Base 0
Option Explicit

Public Type BLOCKARRAY
    Nbytes As Long
    Block() As Byte
End Type

Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (pvDest As Any, lpvSource As Any, ByVal cbCopy As Long)
Private Declare Sub MoveMemory Lib "kernel32" Alias "RtlMoveMemory" (dest As Any, ByVal Source As Long, ByVal length As Long)

Private Declare Function GetTickCount Lib "kernel32" () As Long

Private retval As Long

'============================
'ROUTINES DI INIZIALIZZAZIONE
'============================

Public Sub Inizializza(Blocco As BLOCKARRAY, ByVal Dimensione As Long)
Blocco.Nbytes = Dimensione
ReDim Blocco.Block(Dimensione - 1)
End Sub

Public Sub Riduci(Blocco As BLOCKARRAY)
If Blocco.Nbytes < 2 Then Exit Sub
Dim i As Long
i = Blocco.Nbytes
Do
    i = i - 1
    If Blocco.Block(i) = 0 Then
        Blocco.Nbytes = Blocco.Nbytes - 1
    Else
        Exit Do
    End If
Loop Until i = 1
ReDim Preserve Blocco.Block(Blocco.Nbytes - 1)
End Sub

Public Sub Randomizza(Blocco As BLOCKARRAY)
If Blocco.Nbytes = 0 Then Exit Sub
ReDim Blocco.Block(Blocco.Nbytes - 1)
Dim i As Long
Dim caos As Byte
Randomize GetTickCount
For i = 0 To Blocco.Nbytes - 1
    caos = Int(Rnd * 256)
    Blocco.Block(i) = caos
Next i
End Sub



'======================================
'ROUTINES DI ASSEGNAZIONE E SPOSTAMENTO
'======================================

Public Sub MOV(ByRef BloccoPieno As BLOCKARRAY, BloccoVuoto As BLOCKARRAY)
BloccoVuoto.Nbytes = BloccoPieno.Nbytes
ReDim BloccoVuoto.Block(BloccoPieno.Nbytes - 1)
Dim i As Long
CopyMemory ByVal VarPtr(BloccoVuoto.Block(0)), ByVal VarPtr(BloccoPieno.Block(0)), BloccoPieno.Nbytes
End Sub

Public Sub XCHG(Blocco1 As BLOCKARRAY, Blocco2 As BLOCKARRAY)
Dim Swapblock As BLOCKARRAY
MOV Blocco1, Swapblock
MOV Blocco2, Blocco1
MOV Swapblock, Blocco2
End Sub



'===================
'FUNZIONI DI COMPARE
'===================

Public Function IsZero(ByRef Blocco As BLOCKARRAY) As Boolean
IsZero = False
If Blocco.Nbytes <> 1 Then Riduci Blocco
If Blocco.Nbytes = 1 And Blocco.Block(0) = 0 Then IsZero = True
End Function

Public Function A_uguale_B(ByRef a As BLOCKARRAY, ByRef b As BLOCKARRAY) As Boolean
A_uguale_B = False
If a.Nbytes <> b.Nbytes Then
Riduci a
Riduci b
    If a.Nbytes <> b.Nbytes Then Exit Function
End If
Dim i As Long
For i = 0 To a.Nbytes - 1
    If a.Block(i) <> b.Block(i) Then Exit Function
Next i
A_uguale_B = True
End Function

Public Function A_maggiore_B(ByRef a As BLOCKARRAY, ByRef b As BLOCKARRAY) As Boolean
A_maggiore_B = False
If a.Nbytes <> b.Nbytes Then
    Riduci a
    Riduci b
    If a.Nbytes > b.Nbytes Then A_maggiore_B = True
    If a.Nbytes <> b.Nbytes Then Exit Function
End If
Dim i As Long
Dim Pos As Long
For i = 1 To a.Nbytes
    Pos = a.Nbytes - i
    If a.Block(Pos) <> b.Block(Pos) Then Exit For
Next i
If a.Block(Pos) > b.Block(Pos) Then A_maggiore_B = True
End Function

Public Function A_minore_B(ByRef a As BLOCKARRAY, ByRef b As BLOCKARRAY) As Boolean
A_minore_B = False
If a.Nbytes <> b.Nbytes Then
    Riduci a
    Riduci b
    If a.Nbytes < b.Nbytes Then A_minore_B = True
    If a.Nbytes <> b.Nbytes Then Exit Function
End If
Dim i As Long
Dim Pos As Long
For i = 1 To a.Nbytes
    Pos = a.Nbytes - i
    If a.Block(Pos) <> b.Block(Pos) Then Exit For
Next i
If a.Block(Pos) < b.Block(Pos) Then A_minore_B = True
End Function



'================
'OPERATORI LOGICI
'================

Public Function LogicNOT(ByRef Blocco As BLOCKARRAY) As BLOCKARRAY
LogicNOT.Nbytes = Blocco.Nbytes
ReDim LogicNOT.Block(LogicNOT.Nbytes - 1)
Dim i As Integer
For i = 0 To LogicNOT.Nbytes - 1
    LogicNOT.Block(i) = Not Blocco.Block(i)
Next i
End Function

Public Function LogicAND(ByRef Blocco1 As BLOCKARRAY, ByRef Blocco2 As BLOCKARRAY) As BLOCKARRAY
If Blocco1.Nbytes > Blocco2.Nbytes Then
    ReDim Preserve Blocco2.Block(Blocco1.Nbytes - 1)
    Blocco2.Nbytes = Blocco1.Nbytes
ElseIf Blocco2.Nbytes < Blocco2.Nbytes Then
    ReDim Preserve Blocco1.Block(Blocco2.Nbytes - 1)
    Blocco1.Nbytes = Blocco2.Nbytes
End If
Dim i As Integer
LogicAND.Nbytes = Blocco1.Nbytes
ReDim LogicAND.Block(Blocco1.Nbytes - 1)
For i = 0 To Blocco1.Nbytes - 1
    LogicAND.Block(i) = Blocco1.Block(i) And Blocco2.Block(i)
Next i
Riduci Blocco1
Riduci Blocco2
Riduci LogicAND
End Function

Public Function LogicOR(ByRef Blocco1 As BLOCKARRAY, ByRef Blocco2 As BLOCKARRAY) As BLOCKARRAY
If Blocco1.Nbytes > Blocco2.Nbytes Then
    ReDim Preserve Blocco2.Block(Blocco1.Nbytes - 1)
    Blocco2.Nbytes = Blocco1.Nbytes
ElseIf Blocco2.Nbytes < Blocco2.Nbytes Then
    ReDim Preserve Blocco1.Block(Blocco2.Nbytes - 1)
    Blocco1.Nbytes = Blocco2.Nbytes
End If
Dim i As Integer
LogicOR.Nbytes = Blocco1.Nbytes
ReDim LogicOR.Block(Blocco1.Nbytes - 1)
For i = 0 To Blocco1.Nbytes - 1
    LogicOR.Block(i) = Blocco1.Block(i) Or Blocco2.Block(i)
Next i
Riduci Blocco1
Riduci Blocco2
End Function

Public Function LogicXOR(ByRef Blocco1 As BLOCKARRAY, ByRef Blocco2 As BLOCKARRAY) As BLOCKARRAY
If Blocco1.Nbytes > Blocco2.Nbytes Then
    ReDim Preserve Blocco2.Block(Blocco1.Nbytes - 1)
    Blocco2.Nbytes = Blocco1.Nbytes
ElseIf Blocco1.Nbytes < Blocco2.Nbytes Then
    ReDim Preserve Blocco1.Block(Blocco2.Nbytes - 1)
    Blocco1.Nbytes = Blocco2.Nbytes
End If
Dim i As Integer
LogicXOR.Nbytes = Blocco1.Nbytes
ReDim LogicXOR.Block(Blocco1.Nbytes - 1)
For i = 0 To Blocco1.Nbytes - 1
    LogicXOR.Block(i) = Blocco1.Block(i) Xor Blocco2.Block(i)
Next i
Riduci Blocco1
Riduci Blocco2
Riduci LogicXOR
End Function



'==================
'OPERATORI SUI BYTE
'==================

Public Sub SHL(Blocco As BLOCKARRAY, ByVal Shift As Long)
Dim Swapblock As BLOCKARRAY
Inizializza Swapblock, Blocco.Nbytes + Shift
Dim i As Long
For i = 0 To Shift - 1
    Swapblock.Block(i) = 0
Next i
CopyMemory ByVal VarPtr(Swapblock.Block(Shift)), ByVal VarPtr(Blocco.Block(0)), Blocco.Nbytes
MOV Swapblock, Blocco
End Sub

Public Sub SHR(Blocco As BLOCKARRAY, ByVal Shift As Long)
Dim NewDimension As Long
NewDimension = Blocco.Nbytes - Shift
If NewDimension < 1 Then
    ReDim Blocco.Block(0)
    Blocco.Block(0) = 0
    Blocco.Nbytes = 1
    Exit Sub
End If
MoveMemory ByVal VarPtr(Blocco.Block(0)), ByVal VarPtr(Blocco.Block(Shift)), NewDimension
Blocco.Nbytes = NewDimension
ReDim Preserve Blocco.Block(NewDimension - 1)
End Sub



'===========================
'OPERATORI PARZIALI SEMPLICI
'===========================

Public Function ADD(Blocco As BLOCKARRAY, ByVal valore As Long) As BLOCKARRAY
Dim i As Long
Dim Overflow As Integer
MOV Blocco, ADD
For i = 0 To (Blocco.Nbytes - 1)
    Overflow = Blocco.Block(i) + valore
    Overflow = Overflow And 255
    ADD.Block(i) = Overflow
    If Overflow < Blocco.Block(i) Then
        valore = 1
    Else
        valore = 0
        Exit For
    End If
Next i
If valore <> 0 Then
    ReDim Preserve ADD.Block(Blocco.Nbytes)
    ADD.Block(Blocco.Nbytes) = valore
    ADD.Nbytes = Blocco.Nbytes + 1
End If
End Function

Public Function MUL(ByRef Blocco As BLOCKARRAY, ByVal Coefficiente As Long) As BLOCKARRAY
Dim Overflow As Long
Dim c As Long
Dim a As Integer
Overflow = 0
Inizializza MUL, Blocco.Nbytes
For c = 0 To Blocco.Nbytes - 1
    a = Blocco.Block(c)
    Overflow = a * Coefficiente + Overflow
    MUL.Block(c) = Overflow And 255
    Overflow = Overflow \ 256
Next c
If Overflow <> 0 Then
    ReDim Preserve MUL.Block(MUL.Nbytes)
    MUL.Nbytes = MUL.Nbytes + 1
    MUL.Block(Blocco.Nbytes) = Overflow
End If
End Function



'==============================
'OPERAZIONI MATEMATICHE DI BASE
'==============================

Public Function Addizione(ByRef Blocco1 As BLOCKARRAY, ByRef Blocco2 As BLOCKARRAY, Somma As BLOCKARRAY) As Long
Addizione = 0
If Blocco1.Nbytes > Blocco2.Nbytes Then
    ReDim Preserve Blocco2.Block(Blocco1.Nbytes - 1)
    Somma.Nbytes = Blocco1.Nbytes
Else
    ReDim Preserve Blocco1.Block(Blocco2.Nbytes - 1)
    Somma.Nbytes = Blocco2.Nbytes
End If
Dim Overflow As Integer
Dim i As Integer
Overflow = 0
ReDim Somma.Block(Somma.Nbytes - 1)
For i = 0 To Somma.Nbytes - 1
    Overflow = Overflow + Blocco1.Block(i)
    Overflow = Overflow + Blocco2.Block(i)
    Somma.Block(i) = Overflow And 255
    Overflow = Overflow \ 256
Next i
If Overflow <> 0 Then
    ReDim Preserve Somma.Block(Somma.Nbytes)
    Somma.Block(Somma.Nbytes) = Overflow
    Somma.Nbytes = Somma.Nbytes + 1
End If
ReDim Preserve Blocco1.Block(Blocco1.Nbytes - 1)
ReDim Preserve Blocco2.Block(Blocco2.Nbytes - 1)
End Function

Public Function Sottrazione(ByRef BloccoGrande As BLOCKARRAY, ByRef BloccoPiccolo As BLOCKARRAY, Risultato As BLOCKARRAY) As Long
Sottrazione = 0
If A_minore_B(BloccoGrande, BloccoPiccolo) Then
 Sottrazione = -1
 Exit Function
End If
Dim Swapblock As BLOCKARRAY
Dim park As BLOCKARRAY
MOV BloccoPiccolo, Swapblock
Swapblock.Nbytes = BloccoGrande.Nbytes
ReDim Preserve Swapblock.Block(Swapblock.Nbytes - 1)
Swapblock = LogicNOT(Swapblock)
Swapblock = ADD(Swapblock, 1)
retval = Addizione(BloccoGrande, Swapblock, park)
park.Block(park.Nbytes - 1) = park.Block(park.Nbytes - 1) \ 2
Riduci park
MOV park, Risultato
End Function

Public Function Moltiplicazione(ByRef BloccoGrande As BLOCKARRAY, ByRef BloccoPiccolo As BLOCKARRAY, Prodotto As BLOCKARRAY) As Long
Moltiplicazione = 0
Dim k As Long
Dim m As Long
Dim Overflow As Long
Dim ShiftedMatrix() As Byte
Dim a As Long
Dim b As Long
Dim park As Long
Dim Swapblock As BLOCKARRAY
k = BloccoGrande.Nbytes - 1
m = BloccoPiccolo.Nbytes - 1
Inizializza Prodotto, k + m + 1
ReDim ShiftedMatrix(BloccoGrande.Nbytes, m)
Dim Parkeggio As BLOCKARRAY
Inizializza Parkeggio, BloccoGrande.Nbytes + 1
Inizializza Swapblock, BloccoGrande.Nbytes + 1
'creazione matrice compatta
For b = 0 To m
    Overflow = 0
    For a = 0 To k
        park = BloccoGrande.Block(a)
        park = park * BloccoPiccolo.Block(b)
        Overflow = park + Overflow
        ShiftedMatrix(a, b) = Overflow And 255
        Overflow = Overflow \ 256
    Next a
    ShiftedMatrix(BloccoGrande.Nbytes, b) = Overflow
Next b
'shifting blocchi e addizione matrice
For b = 0 To m
    For a = 0 To BloccoGrande.Nbytes
        Parkeggio.Block(a) = ShiftedMatrix(a, m - b)
    Next a
    SHL Prodotto, 1
    MOV Prodotto, Swapblock
    retval = Addizione(Parkeggio, Swapblock, Prodotto)
Next b
If Prodotto.Block(Prodotto.Nbytes - 1) = 0 Then Riduci Prodotto
End Function

Public Function Divisione(ByRef Dividendo As BLOCKARRAY, ByRef Divisore As BLOCKARRAY, Quoziente As BLOCKARRAY, Resto As BLOCKARRAY) As Long
'Algoritmo di divisione
Divisione = 0
On Error GoTo 503
Inizializza Quoziente, 1
Inizializza Resto, 1
Dim park As BLOCKARRAY
Dim Prova As BLOCKARRAY
Inizializza park, 1
Inizializza Prova, 1
Dim i As Long
Dim Kdendo As Long
Dim Ksore As Long
Dim cont As Long
Dim cifra As Long
Dim Coefficiente As Long
Kdendo = Dividendo.Nbytes - 1
Ksore = Divisore.Nbytes - 1
'START
'controlla i casi banali
If A_minore_B(Dividendo, Divisore) Then
    MOV Dividendo, Resto
    Exit Function
End If
If A_uguale_B(Dividendo, Divisore) Then
    Quoziente.Block(0) = 1
    Exit Function
End If
If IsZero(Divisore) Then
    Divisione = 11
    Exit Function
End If
If IsZero(Dividendo) Then
    Exit Function
End If
'nessun caso banale: esegue l'algoritmo
'INIZIALIZZAZIONE
Inizializza Resto, Divisore.Nbytes
For i = 0 To Ksore
    Resto.Block(i) = Dividendo.Block(Kdendo - Ksore + i)
Next i
cont = Divisore.Nbytes
'DIVISIONE PRELIMINARE
cifra = Resto.Block(Resto.Nbytes - 1)
If A_minore_B(Resto, Divisore) Then
    SHL Resto, 1
    Resto.Block(0) = Dividendo.Block(Kdendo - cont)
    cont = cont + 1
    cifra = cifra * 256
    cifra = cifra + Resto.Block(Resto.Nbytes - 2)
End If
Coefficiente = cifra \ Divisore.Block(Ksore)
666:
Prova = MUL(Divisore, Coefficiente)
If A_minore_B(Resto, Prova) Then
    Coefficiente = Coefficiente - 1
    GoTo 666
End If
Quoziente.Block(0) = Coefficiente
retval = Sottrazione(Resto, Prova, Resto)
If cont = Dividendo.Nbytes Then Exit Function
'INIZIO CICLO
Do
    SHL Resto, 1
    Resto.Block(0) = Dividendo.Block(Kdendo - cont)
    cont = cont + 1
    cifra = Resto.Block(Resto.Nbytes - 1)
    If Resto.Nbytes > Divisore.Nbytes Then
        cifra = cifra * 256
        cifra = cifra + Resto.Block(Resto.Nbytes - 2)
    End If
    Coefficiente = cifra \ Divisore.Block(Ksore)
667:
    Prova = MUL(Divisore, Coefficiente)
    If A_minore_B(Resto, Prova) Then
        Coefficiente = Coefficiente - 1
        GoTo 667
    End If
    SHL Quoziente, 1
    Quoziente.Block(0) = Coefficiente
    retval = Sottrazione(Resto, Prova, Resto)
Loop Until cont = Dividendo.Nbytes
'fine
Exit Function
503:
Divisione = Err.Number
On Error GoTo 0
End Function



'===================
'OPERAZIONI AVANZATE
'===================

Public Function ModuloEsponenziale(ByRef Base As BLOCKARRAY, ByRef Esponente As BLOCKARRAY, ByRef Modulo As BLOCKARRAY, Resto As BLOCKARRAY) As Long
ModuloEsponenziale = 0
On Error GoTo 21011
Dim Quozpark As BLOCKARRAY
Inizializza Quozpark, 1
Inizializza Resto, 1
retval = Divisione(Base, Modulo, Quozpark, Resto)
If retval <> 0 Then
    ModuloEsponenziale = retval
    Exit Function
End If
Dim Superresto As BLOCKARRAY
Inizializza Superresto, 1
Superresto.Block(0) = 1
Dim park As BLOCKARRAY
MOV Resto, park
Dim Matrice As Integer
Dim i As Long
Dim j As Long
For i = 0 To (Esponente.Nbytes - 1)
    Matrice = 1
    For j = 0 To 7
        If (Esponente.Block(i) And Matrice) <> 0 Then
            retval = Moltiplicazione(Superresto, park, Resto)
            retval = Divisione(Resto, Modulo, Quozpark, Superresto)
        End If
        retval = Moltiplicazione(park, park, Resto)
        retval = Divisione(Resto, Modulo, Quozpark, park)
        Matrice = Matrice * 2
    Next j
Next i
MOV Superresto, Resto
Exit Function
21011:
ModuloEsponenziale = Err.Number
End Function

Public Function Elevamento(ByRef Base As BLOCKARRAY, ByRef Esponente As BLOCKARRAY, Potenza As BLOCKARRAY) As Long
Elevamento = 0
On Error GoTo 21011
Dim Swapblock As BLOCKARRAY
Inizializza Swapblock, 1
Swapblock.Block(0) = 1
Inizializza Potenza, 1
Potenza.Block(0) = 1
Dim Pork As BLOCKARRAY
Dim park As BLOCKARRAY
MOV Base, park
Dim Matrice As Integer
Dim i As Long
Dim j As Long
For i = 0 To (Esponente.Nbytes - 1)
    Matrice = 1
    For j = 0 To 7
        If (Esponente.Block(i) And Matrice) <> 0 Then
            retval = Moltiplicazione(Potenza, park, Pork)
            MOV Pork, Potenza
        End If
        retval = Moltiplicazione(park, park, Swapblock)
        MOV Swapblock, park
        Matrice = Matrice * 2
    Next j
Next i
Exit Function
21011:
Elevamento = Err.Number
End Function

Public Function InversoMoltiplicativo(ByRef Blocco As BLOCKARRAY, ByRef Modulo As BLOCKARRAY, Inverso As BLOCKARRAY) As Long
'alias Extended Euclidean Algorithm
InversoMoltiplicativo = 0
On Error GoTo 2327
Dim a As BLOCKARRAY
Dim b As BLOCKARRAY
Dim t0 As BLOCKARRAY
Dim c As BLOCKARRAY
Dim q As BLOCKARRAY
Dim park As BLOCKARRAY
Dim park2 As BLOCKARRAY
MOV Modulo, a
MOV Blocco, b
Inizializza Inverso, 1
Inverso.Block(0) = 1
If b.Nbytes = 1 And b.Block(0) = 1 Then Exit Function
retval = Divisione(a, b, t0, c)
Do Until (c.Nbytes = 1 And c.Block(0) = 1)
    retval = Divisione(b, c, q, park)
    MOV park, b
    retval = Moltiplicazione(q, t0, park)
    MOV Inverso, park2
    retval = Addizione(park, park2, Inverso)
    If b.Nbytes = 1 And b.Block(0) = 1 Then Exit Function
    retval = Divisione(c, b, q, park)
    MOV park, c
    retval = Moltiplicazione(q, Inverso, park)
    MOV t0, park2
    retval = Addizione(park, park2, t0)
Loop
retval = Sottrazione(a, t0, Inverso)
Exit Function
2327:
InversoMoltiplicativo = Err.Number
End Function

Public Function Differenza(ByRef Blocco1 As BLOCKARRAY, ByRef Blocco2 As BLOCKARRAY) As BLOCKARRAY
If A_uguale_B(Blocco1, Blocco2) Then
    Inizializza Differenza, 1
    Differenza.Block(0) = 0
    Exit Function
End If
If A_maggiore_B(Blocco1, Blocco2) Then
    retval = Sottrazione(Blocco1, Blocco2, Differenza)
Else
    retval = Sottrazione(Blocco2, Blocco1, Differenza)
End If
End Function

Public Function SottrazioneModulare(ByRef Blocco1 As BLOCKARRAY, ByRef Blocco2 As BLOCKARRAY, ByRef Modulo As BLOCKARRAY, Risultato As BLOCKARRAY) As Long
SottrazioneModulare = 0
On Error GoTo 9652
If IsZero(Modulo) Then Err.Raise -1
Dim Quozpark As BLOCKARRAY
Dim park As BLOCKARRAY
Risultato = Differenza(Blocco1, Blocco2)
retval = Divisione(Risultato, Modulo, Quozpark, park)
If A_maggiore_B(Blocco1, Blocco2) Then
    MOV park, Risultato
Else
    retval = Sottrazione(Modulo, park, Risultato)
End If
Exit Function
9652:
SottrazioneModulare = Err.Number
End Function

Public Function DivisioneModulare(ByRef Dividendo As BLOCKARRAY, ByRef Divisore As BLOCKARRAY, ByRef Modulo As BLOCKARRAY, Quoziente As BLOCKARRAY) As Long
DivisioneModulare = 0
On Error GoTo 557
Dim park As BLOCKARRAY
Dim Temp As BLOCKARRAY
retval = InversoMoltiplicativo(Divisore, Modulo, park)
If retval <> 0 Then Err.Raise retval
retval = Moltiplicazione(Dividendo, park, Temp)
retval = Divisione(Temp, Modulo, park, Quoziente)
If retval <> 0 Then Err.Raise retval
Exit Function
557:
DivisioneModulare = Err.Number
End Function



'=================
'FUNZIONI DI INPUT
'=================

Public Function NumericStringToBlock(ByVal Number As String, Blocco As BLOCKARRAY) As Long
NumericStringToBlock = 0
On Error GoTo 4000
Dim Lunghezza As Long
Dim cifra As Long
Dim p As Long
Lunghezza = Len(Number)
Inizializza Blocco, 1
For p = 1 To Lunghezza
    cifra = Asc(Mid(Number, p, 1))
    If cifra < 48 Or cifra > 57 Then
        NumericStringToBlock = -2
        Exit Function
    End If
    cifra = cifra Xor 48
    Blocco = MUL(Blocco, 10)
    Blocco = ADD(Blocco, cifra)
Next p
Exit Function
4000:
NumericStringToBlock = Err.Number
End Function



'==================
'FUNZIONI DI OUTPUT
'==================

Public Function BlockToNumericString(ByRef Blocco As BLOCKARRAY) As String
BlockToNumericString = Empty
If Blocco.Nbytes < 1 Then Exit Function
Dim i As Long
Dim Resto As BLOCKARRAY
Dim park As BLOCKARRAY
Dim Quoziente As BLOCKARRAY
Dim Dieci As BLOCKARRAY
MOV Blocco, park
Inizializza Dieci, 1
Dieci.Block(0) = 10
Do
    retval = Divisione(Blocco, Dieci, Quoziente, Resto)
    BlockToNumericString = Trim(Str(Resto.Block(0))) & BlockToNumericString
    MOV Quoziente, Blocco
Loop Until IsZero(Quoziente)
MOV park, Blocco
End Function

Public Function BlockToAsciiString(ByRef Blocco As BLOCKARRAY) As String
BlockToAsciiString = Empty
If Blocco.Nbytes < 1 Then Exit Function
Dim i As Long
Dim carattere As String * 1
For i = 0 To Blocco.Nbytes - 1
    carattere = Chr(Blocco.Block(i))
    BlockToAsciiString = carattere & BlockToAsciiString
Next i
End Function





'=====
'VARIE
'=====

Public Function BitLenght(ByRef Blocco As BLOCKARRAY) As Long
BitLenght = 0
If Blocco.Nbytes <= 0 Then Exit Function
Dim Swapblock As BLOCKARRAY
Dim x As Byte
MOV Blocco, Swapblock
Riduci Swapblock
x = Swapblock.Block(Swapblock.Nbytes - 1)
Do
    BitLenght = BitLenght + 1
    x = x \ 2
Loop Until x = 0
BitLenght = BitLenght + (Swapblock.Nbytes - 1) * 8
End Function

Public Function StringToHex(ByVal ASCIItext As String, ByRef hexBuffer As String) As Long
StringToHex = 0
'DICHIARAZIONI
Dim i As Long
Dim digit As String
Dim car As Byte
Dim Lunghezza As Long
Dim Pos As Long
'INIZIALIZZAZIONE
Lunghezza = Len(ASCIItext)
hexBuffer = String(Lunghezza * 2, "0")
Pos = 1
'CODIFICA
For i = 1 To Lunghezza
    car = Asc(Mid(ASCIItext, i, 1))
    digit = Hex(car)
    If car < 16 Then
        'valore composto da una sola cifra
        'esadecimale: salta uno "0"
        Pos = Pos + 1
        Mid(hexBuffer, Pos, 1) = digit
        'scorre al prossimo valore
        Pos = Pos + 1
    Else
        Mid(hexBuffer, Pos, 2) = digit
        'scorre al prossimo valore
        Pos = Pos + 2
    End If
Next i
End Function

Public Function RabinMillerTest(ByRef Blocco As BLOCKARRAY, ByVal Tentativi As Long) As Boolean
RabinMillerTest = False
If Tentativi <= 0 Then Exit Function
Dim Swapblock As BLOCKARRAY
Dim Uno As BLOCKARRAY
Dim DUE As BLOCKARRAY
Dim park As BLOCKARRAY
Dim Resto As BLOCKARRAY
Dim Temp As BLOCKARRAY
Inizializza Uno, 1
Uno.Block(0) = 1
Inizializza DUE, 1
DUE.Block(0) = 2
retval = Sottrazione(Blocco, Uno, Swapblock)
MOV Swapblock, Temp
Dim Matrice As Integer
Dim Contazeri As Long
Contazeri = 0
Dim i As Long
Dim j As Long
For i = 0 To Swapblock.Nbytes - 1
    Matrice = 1
    For j = 0 To 7
        If (Swapblock.Block(i) And Matrice) = 0 Then
            Contazeri = Contazeri + 1
        Else
            GoTo out
        End If
        Matrice = Matrice * 2
    Next j
Next i
out:
Dim Scarto As Long
Scarto = Contazeri And 7
If (Contazeri \ 8) <> 0 Then SHL Swapblock, Contazeri \ 8
Uno.Block(0) = 2 ^ Scarto
retval = Divisione(Swapblock, Uno, park, Resto)
Uno.Block(0) = 1
'MAIN LOOP (numero controlli)
'Inizializza Swapblock, Blocco.Nbytes - 1
For i = 1 To Tentativi
    Inizializza Swapblock, 2
    Randomizza Swapblock
    retval = ModuloEsponenziale(Swapblock, park, Blocco, Resto)
    If (Resto.Nbytes = 1) And Resto.Block(0) = 1 Then GoTo Vittoria
    If A_uguale_B(Resto, Temp) Then GoTo Vittoria
    'INTERNAL LOOP (random)
    For j = 1 To Contazeri
        If (Resto.Nbytes = 1) And Resto.Block(0) = 1 Then Exit Function
        If A_uguale_B(Resto, Temp) Then GoTo Vittoria
        retval = ModuloEsponenziale(Resto, DUE, Blocco, Resto)
    Next j
    Exit Function
'CONTROLLO SUPERATO
Vittoria:
Next i
'TEST RIUSCITO
RabinMillerTest = True
End Function
